/* Copyright (C) 1989 Aladdin Enterprises.  All rights reserved.
   Distributed by Free Software Foundation, Inc.

This file is part of Ghostscript.

Ghostscript is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility
to anyone for the consequences of using it or for whether it serves any
particular purpose or works at all, unless he says so in writing.  Refer
to the Ghostscript General Public License for full details.

Everyone is granted permission to copy, modify and redistribute
Ghostscript, but only under the conditions described in the Ghostscript
General Public License.  A copy of this license is supposed to have been
given to you along with Ghostscript so you can know your rights and
responsibilities.  It should be in a file named COPYING.  Among other
things, the copyright notice and this notice must be preserved on all
copies.  */

/* gsmain.c */
/* Framework for GhostScript drivers */
# ifdef __MSDOS__
#  include <fcntl.h>
# endif
#include <string.h>
#include "gx.h"
#include "malloc_.h"

/*
 * This routine provides the following standard services for parsing
 * a command line:
 *	- setting debug flags (-Z switch) [if debugging];
 *	- tracing (-T switch) [if debugging];
 *	- selecting device (__MSDOS__ flag)
 *	- passing other arguments and switches back to the caller
 *
 * Calling convention:
 *	gs_main(argc, argv, map_name, switch_proc, arg_proc)
 * Calls
 *	switch_proc(switch_char, rest_of_arg) for switches,
 *	arg_proc(arg, index) for non-switch args.
 * If switch_proc returns a negative value, gs_main prints an
 *   "unknown switch" error message and aborts.
 * gs_main returns the number of non-switch args handed to arg_proc.
 */

static unsigned long proc_reloc;	/* relocation of procedures */

int
gs_main(int argc, char *argv[], char *map_name,
    int (*switch_proc)(P2(char, char *)),
    void (*arg_proc)(P2(char *, int)))
{	int argi = 1;
	FILE *mapf;
	int arg_count = 0;
	proc_reloc = 0;
#ifdef gs_DEBUG
	/* Reset debugging flags */
	memset(gs_debug, 0, 128);
	/* Open the map file and look up 'main' */
	mapf = fopen(map_name, "rt");
	if ( mapf == NULL )
		printf("Unable to open map file %s\n", map_name);
	else
	   {	extern char *trace_find_symbol(P2(char *, FILE *));
		char *main_addr = trace_find_symbol("_MAIN", mapf);
		if ( main_addr == NULL )
			printf("Unable to find _main in map file\n");
		else
		   {	extern main();
			proc_reloc = (unsigned long)main - (unsigned long)main_addr;
		   }
	   }
#endif
	/* Initialize the proper device */
	   {	typedef struct gx_device_s gx_device;
		gx_device *dev;
		extern gx_device *gx_device_default_p;
#ifdef __MSDOS__
		/* Install the EGA device */
		extern gx_device ega_device;
		dev = &ega_device;
#else
		/* Install the X Windows device */
		extern gx_device x_device;
		dev = &x_device;
#endif
		/* If debugging is enabled, trace the device calls. */
#ifdef gs_DEBUG
		   {	extern gx_device *gs_trace_device(P1(gx_device *));
			gx_device *tdev = gs_trace_device(dev);
			if ( tdev == 0 )
			   {	printf("Can't allocate traced device!");
				abort();
			   }
			dev = tdev;
		   }
#endif
		gx_device_default_p = dev;
	   }
#ifdef __MSDOS__
	/* Open files in 'binary' mode */
	_fmode = O_BINARY;
#endif
	for ( ; argi < argc; argi++ )
	   {	char *arg = argv[argi];
		if ( *arg == '-' )
		   {	switch ( *++arg )
			   {
			default:
				if ( (*switch_proc)(*arg, arg + 1) < 0 )
				   {	printf("Unknown switch %s", arg - 1);
					abort();
				   }
				break;
			case 'Z':
#ifdef gs_DEBUG
				if ( !arg[1] )
				   {	/* No options, set all flags */
					memset(gs_debug, -1, 128);
				   }
				else
				   {	while ( *++arg )
						gs_debug[*arg & 127] = -1;
				   }
#else
				printf("Not a debugging configuration, -Z switch ignored\n");
#endif
				break;
			case 'T':
#ifdef gs_DEBUG
			if ( mapf != NULL )
			   {	char *delim;
				char *tname;
				int rsize = 0;
				extern int trace_flush_flag;
				delim = strchr(arg, ':');
				if ( delim != NULL )
				   {	sscanf(delim + 1, "%d", &rsize);
					*delim = 0;	/* terminate name */
				   }
				tname = malloc(strlen(arg) + 1);
				strcpy(tname, arg);
				*tname = '_';
				strupr(tname);
				if ( trace_name(tname, mapf, NULL, rsize) < 0 )
					printf("%s not found\n", tname);
				trace_flush_flag = 1;
			   }
#else
				printf("Not a debugging configuration, -T switch ignored\n");
#endif
				break;
			   }
		   }
		else
			(*arg_proc)(arg, arg_count++);
	   }
	return arg_count;
}

/* ------ Debugging routines ------ */

#ifdef gs_DEBUG			/* ------ ------ */

void gs_dump_C_stack();

/* Log an error return */
int
gs_log_error(int err)
{	if ( gs_debug['e'] )
	  { printf("Returning error %d:\n", err);
	    gs_dump_C_stack();
	  }
	return err;
}

#endif				/* ------ ifdef gs_DEBUG ------ */

/* Dump the C stack. */
/* This is actually only used for debugging. */
extern char *stack_top_frame();
extern unsigned long stack_return(P1(char *));
extern char *stack_next_frame(P1(char *));
void
gs_dump_C_stack()
{	char *nbp = stack_top_frame();
	char *bp;
	do
	   {	bp = nbp;
		printf("frame %8lx called from %8lx (%8lx)\n",
			(unsigned long)bp, stack_return(bp),
			stack_return(bp) - proc_reloc);
		nbp = stack_next_frame(bp);
	   }
	while ( nbp != 0 );
}
